// Expressive errors and colors
const chalk = require('chalk');
// Chokidar for watching file changes
const chokidar = require('chokidar');
// File pattern globs
const globby = require('globby');
// File system + extra utils
const fs = require('fs-extra');
// Path handling utilities
const path = require('path');

const Observable = require('@runforest/run/utils/observable');

/**
@typedef {{
  source: string;
  destination: string;
  globPattern: string[];
  mode: 'build' | 'watch';
}} ValidatedConfig
@typedef {Partial<Omit<ValidatedConfig, 'mode'> & { mode: string; }>} Config
 */

/**
 * Returns Promise.resolve with validated config or Promise.reject if config is
 * not valid.
 *
 * @param {Config} config Config recived from `@runforest/run`.
 * @return {Promise<ValidatedConfig>}
 */
const validate = (config) =>
  new Promise((resolve, reject) => {
    let { source = 'src/html/', destination = 'public/', globPattern = [] } = config;
    const { mode = 'build' } = config;

    if (!source) {
      reject(new Error('Source is required!'));
      return;
    }

    if (!destination) {
      reject(new Error('Destination is required!'));
      return;
    }

    if (!globPattern) {
      reject(new Error('Glob pattern is required!'));
      return;
    }

    source = path.resolve(source);
    destination = path.resolve(destination);

    globPattern = [
      // Include everything inside html folder.
      '**/*',
      // Directories to be excluded from the copy.
      '!sass',
      '!sass/**',
      '!js',
      '!js/**',
      '!svg',
      '!svg/**',
    ].concat(globPattern);

    resolve({
      source,
      destination,
      globPattern,
      mode: mode !== 'watch' ? 'build' : 'watch',
    });
  });

/**
 * Copies files from source to destination.
 *
 * @param  {ValidatedConfig}   config   Config recived from `@runforest/run`.
 * @param  {import('@runforest/run/utils/observable').Observer<any>} observer Task observer.
 * @return {Promise<ValidatedConfig>}
 */
const build = async (config, observer) => {
  const { source, destination, globPattern } = config;

  observer.next(`Copying files from ${source} to ${destination} ...`);

  const paths = await globby(globPattern, { cwd: source });
  const items = await Promise.all(
    paths.map((filePath) => fs.copy(path.resolve(source, filePath), path.resolve(destination, filePath)))
  );

  observer.next(chalk`{green Copied ${items.length} items}`);
  return config;
};

/**
 * Sync files from source to destination.
 *
 * @param  {ValidatedConfig}   config   Config recived from `@runforest/run`.
 * @param  {import('@runforest/run/utils/observable').Observer<any>} observer Task observer.
 * @return {Promise<ValidatedConfig>}
 */
const watch = (config, observer) => {
  const { source, destination, globPattern } = config;

  observer.next(`Syncing files from ${source} to ${destination} ...`);

  return new Promise((resolve, reject) => {
    // watch for file changes and sync them over
    const watcher = chokidar.watch(
      // Filter out exclusion glob patterns.
      globPattern,
      {
        // Ignore dot-files.
        ignored: [/(^|[/\\])\../],
        cwd: source,
      }
    );

    /**
     * Copies file or directory from source to public.
     *
     * @param {string} changedFilePath File or directory source path.
     * @return {void}
     */
    const copyFn = (changedFilePath) => {
      const sourcePath = path.resolve(source, changedFilePath);
      const destinationPath = path.resolve(destination, changedFilePath);

      observer.next(`Copying ${sourcePath} to ${destinationPath} ...`);

      const paths = globby.sync([changedFilePath].concat(globPattern.filter((pattern) => pattern.indexOf('!') === 0)), {
        cwd: source,
      });

      for (let i = 0; i < paths.length; i += 1) {
        const filePath = paths[i];

        fs.copySync(path.resolve(source, filePath), path.resolve(destination, filePath));
      }

      observer.next(`Copied ${sourcePath} to ${destinationPath}`);
    };

    /**
     * Removes file or directory from public.
     *
     * @param {string} filePath File or dir source path.
     * @return {void}
     */
    const unlinkFn = (filePath) => {
      const destinationPath = path.resolve(destination, filePath);

      observer.next(`Removing ${destinationPath} ...`);

      fs.removeSync(destinationPath);

      observer.next(`Removed ${destinationPath}`);
    };

    watcher
      // Copy a single file
      .on('add', copyFn)
      // Recursively copy directory
      .on('addDir', copyFn)
      // On unlink dir, remove the file recursively
      .on('unlink', unlinkFn)
      // On unlink dir, remove the file recursively
      .on('unlinkDir', unlinkFn)
      .on('change', copyFn)
      .on('error', (error) => {
        watcher.close();

        observer.next(chalk`{red Unhandled watcher exception}`);
        reject(error);
      });
  });
};

/**
 * @param {Config} config
 */
module.exports = (config) =>
  new Observable((observer) => {
    validate(config)
      .then((validatedConfig) => {
        const { mode } = validatedConfig;

        if (mode === 'watch') {
          return watch(validatedConfig, observer);
        }

        return build(validatedConfig, observer);
      })
      .then(() => {
        observer.complete();
      })
      .catch((error) => {
        observer.error(error);
      });
  });
